package com.blazinggames.urcl.ui.playfield;

// Java imports
import java.awt.*;

/**
 * Ultimate Retro Common Library
 * Copyright (C) 2001, 2003-2004 Blazing Games Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ------------------------------------------------
 * 
 * A playfield is a grid of tiles that make up a display.
 * This is a very basic playfield class, and unfortunatly is on the
 * slow side. For an arcade game you would be best off either creating
 * your own raw renderer using an image producer.
 * 
 * @author Billy D. Spelchan
 * @version OSR1.00
 */
public class Playfield
{
// *********************
// ***** VARIABLES *****
// *********************
    protected int _width;
    protected int _height;

    protected int _numTileTypes;
    protected int _tileWidth;
    protected int _tileHeight;
    
    // Tile and attributes stored as [row][column]
    protected short _tileData[][];
    protected short _attributeData[][];
    protected TileRenderer _tileset[];


// ************************
// ***** CONSTRUCTORs *****
// ************************

    /**
     * Constructor.
     */
    public Playfield()
    {
        _width = 0;
        _height = 0;
        _tileData = null;
        _attributeData = null;
        _tileset = null;
    }


// ****************************
// ***** METHODS - SETUP! *****
// ****************************

    /**
     * sets the number and size of the tiles that the playfield uses.
     *
     * @param n number of different tiles in the tileset
     * @param w width of each tile
     * @param h height of each tile
    */
    public void setTileset(int n, int w, int h)
    {
        _numTileTypes = n;
        _tileWidth = w;
        _tileHeight = h;
        _tileset = new TileRenderer[n];
    }

    /**
     * Sets the tilerenderer to use for rendering the indicated tileset slot.
     *
     * @param slot which tileset slot to assign the renderer to
     * @param tr the renderer to render the tile with
    */
    public void setTileRenderer(int slot, TileRenderer tr)
    {
        _tileset[slot] = tr;
    }

    /**
     * sets the size of the playfield.
     *
     * @param w width of the playfield
     * @param h height of the playfield
    */
    public void setPlayfield(int w, int h)
    {
        _width = w;
        _height = h;
        _tileData = new short[h][w];
        _attributeData = new short[h][w];
    }


// ********************************************
// ***** METHODS - Playfield manipulation *****
// ********************************************

    /**
     * Clears the playfield using the indicated tile id and the indicated
     * attribute for the tile.
     *
     * @param tile tile id to set all the tiles to
     * @param attribute tile attribute to set all attributes to
     */
    public void clear(int tile, int attribute)
    {
        short st = (short) tile;
        short sa = (short) attribute;
        for (int cntrRow = 0; cntrRow < _height; ++cntrRow)
        {
            for (int cntrCol = 0; cntrCol < _width; ++cntrCol)
            {
                _tileData[cntrRow][cntrCol] = st;
                _attributeData[cntrRow][cntrCol] = sa;
            }
        }
    }

    /**
     * shifts the tiles in the indicated direction
     *
     * @param x how much to scroll the display in the x direction
     * @param y how much to scroll the display in the y direction
    */
    public synchronized void scroll(int x, int y)
    {
        int yShift = Math.abs(y) % _height;
        int cntrRow, cntrCol;
        
        // first phase is to scroll up/down
        if (yShift > 0)
        { 
            short copyT[][] = new short[yShift][];
            short copyA[][] = new short[yShift][];
            if (y > 0)
            {
                for (cntrRow = 0; cntrRow < yShift; ++cntrRow)
                {
                    copyT[cntrRow] = _tileData[_height - yShift + cntrRow];
                    copyA[cntrRow] = _attributeData[_height - yShift + cntrRow];
                }
                for (cntrRow = _height-1; cntrRow >= yShift; --cntrRow)
                {
                    _tileData[cntrRow] = _tileData[cntrRow-yShift];
                    _attributeData[cntrRow] = _attributeData[cntrRow-yShift];
                }
                for (cntrRow = 0; cntrRow < yShift; ++cntrRow)
                {
                    _tileData[cntrRow] = copyT[cntrRow];
                    _attributeData[cntrRow] = copyA[cntrRow];
                }
            }
            else
            {
                for (cntrRow = 0; cntrRow < yShift; ++cntrRow)
                {
                    copyT[cntrRow] = _tileData[cntrRow];
                    copyA[cntrRow] = _attributeData[cntrRow];
                }
                for (cntrRow = yShift; cntrRow < _height; ++cntrRow)
                {
                    _tileData[cntrRow-yShift] = _tileData[cntrRow];
                    _attributeData[cntrRow-yShift] = _attributeData[cntrRow];
                }
                for (cntrRow = 0; cntrRow < yShift; ++cntrRow)
                {
                    _tileData[_height - yShift + cntrRow] = copyT[cntrRow];
                    _attributeData[_height - yShift + cntrRow] = copyA[cntrRow];
                }
            }
        }
        
        // second phase - shift the columns
        int xShift = Math.abs(x) % _width;
        if (xShift > 0)
        {
            short copyT[] = new short[xShift];
            short copyA[] = new short[xShift];
            for (cntrRow = 0; cntrRow < _height; ++cntrRow)
            {
                if (x < 0)
                {
                    System.arraycopy(_tileData[cntrRow], 0, copyT, 0, xShift);
                    System.arraycopy(_attributeData[cntrRow], 0, copyA, 0, xShift);
                    
                    System.arraycopy(_tileData[cntrRow], xShift, _tileData[cntrRow], 0, _width - xShift);
                    System.arraycopy(_attributeData[cntrRow], xShift, _attributeData[cntrRow], 0, _width - xShift);
                    
                    System.arraycopy(copyT, 0, _tileData[cntrRow], _width - xShift, xShift);
                    System.arraycopy(copyA, 0, _attributeData[cntrRow], _width - xShift, xShift);
                }
                else
                {
                    System.arraycopy(_tileData[cntrRow], _width - xShift, copyT, 0, xShift);
                    System.arraycopy(_attributeData[cntrRow], _width - xShift, copyA, 0, xShift);
                    
                    System.arraycopy(_tileData[cntrRow], 0, _tileData[cntrRow], xShift, _width - xShift);
                    System.arraycopy(_attributeData[cntrRow], 0, _attributeData[cntrRow], xShift, _width - xShift);
                    
                    System.arraycopy(copyT, 0, _tileData[cntrRow], 0, xShift);
                    System.arraycopy(copyA, 0, _attributeData[cntrRow], 0, xShift);
                }
            }
        }        
    }

    /**
     * Shifts the display replacing the shifted area with the indicated
     * tile using the indicated attribute.
     *
     * @param x how much to scroll the display in the x direction
     * @param y how much to scroll the display in the y direction
     * @param tile tile id to set all the tiles to
     * @param attr tile attribute to set all attributes to
    */
    public synchronized void shift(int x, int y, short tile, short attr)
    {
        int yShift = Math.abs(y) % _height;
        int cntrRow, cntrCol;
        
        // first phase is to shift up/down (taking into account that we are copying array references)
        if (yShift > 0)
        { 
            if (y > 0)
            {
                for (cntrRow = _height-1; cntrRow >= yShift; --cntrRow)
                {
                    _tileData[cntrRow] = _tileData[cntrRow-yShift];
                    _attributeData[cntrRow] = _attributeData[cntrRow-yShift];
                }
                for (cntrRow = 0; cntrRow < yShift; ++cntrRow)
                {
                    _tileData[cntrRow] = new short[_width];
                    _attributeData[cntrRow] = new short[_width];
                    for (cntrCol = 0; cntrCol < _width; ++cntrCol)
                    {
                        _tileData[cntrRow][cntrCol] = tile;
                        _attributeData[cntrRow][cntrCol] = attr; 
                    }
                }
            }
            else
            {
                for (cntrRow = yShift; cntrRow < _height; ++cntrRow)
                {
                    _tileData[cntrRow-yShift] = _tileData[cntrRow];
                    _attributeData[cntrRow-yShift] = _attributeData[cntrRow];
                }
                for (cntrRow = 0; cntrRow < yShift; ++cntrRow)
                {
                    _tileData[_height - yShift + cntrRow] = new short[_width];
                    _attributeData[_height - yShift + cntrRow] = new short[_width];
                    for (cntrCol = 0; cntrCol < _width; ++cntrCol)
                    {
                        _tileData[_height - yShift + cntrRow][cntrCol] = tile;
                        _attributeData[_height - yShift + cntrRow][cntrCol] = attr; 
                    }
                }
            }
        }
        
        // second phase - shift the columns
        int xShift = Math.abs(x) % _width;
        if (xShift > 0)
        {
            for (cntrRow = 0; cntrRow < _height; ++cntrRow)
            {
                if (x < 0)
                {
                    System.arraycopy(_tileData[cntrRow], xShift, _tileData[cntrRow], 0, _width - xShift);
                    System.arraycopy(_attributeData[cntrRow], xShift, _attributeData[cntrRow], 0, _width - xShift);
                    
                    for (cntrCol = 0; cntrCol < _width; ++cntrCol)
                    {
                        _tileData[_width - yShift + cntrRow][cntrCol] = tile;
                        _attributeData[_width - yShift + cntrRow][cntrCol] = attr; 
                    }
                }
                else
                {
                    System.arraycopy(_tileData[cntrRow], 0, _tileData[cntrRow], xShift, _width - xShift);
                    System.arraycopy(_attributeData[cntrRow], 0, _attributeData[cntrRow], xShift, _width - xShift);
                    
                    for (cntrCol = 0; cntrCol < _width; ++cntrCol)
                    {
                        _tileData[cntrRow][cntrCol] = tile;
                        _attributeData[cntrRow][cntrCol] = attr; 
                    }
                }
            }
        }
    }


    /**
     * sets the indicated tile to the indicated value leaving the
     * attribute the way it is.
     *
     * @param x playfield x coordinate
     * @param y playfield y coordinate
     * @param tile tile value to use
     */
    public void setTile(int x, int y, int tile)
    {
        _tileData[y][x] = (short) tile;
    }

    /**
     * sets the indicated tile to the indicated value as well
     * as the attribute.
     *
     * @param x playfield x coordinate
     * @param y playfield y coordinate
     * @param tile tile value to use
     * @param attribute tile attribute value to use
     */
    public void setTile(int x, int y, int tile, int attr)
    {
        _tileData[y][x] = (short) tile;
        _attributeData[y][x] = (short) attr;
    }

    /**
     * sets the indicated tile to the indicated value leaving the
     * attribute the way it is.
     *
     * @param x playfield x coordinate
     * @param y playfield y coordinate
     * @param attribute tile attribute value to use
     */
    public void setAttribute(int x, int y, int attr)
    {
        _attributeData[y][x] = (short) attr;
    }
    

// *********************************
// ***** METHODS - Information *****
// *********************************

    /**
     * returns the value of the tile at the indicated location
     *
     * @param x playfield x coordinate
     * @param y playfield y coordinate
     * @return the value of the tile at the indicated location
    */
    public int getTile(int x, int y)
    {
        return _tileData[y][x];
    }

    /**
     * returns the value of the attribute at the indicated location
     *
     * @param x playfield x coordinate
     * @param y playfield y coordinate
     * @return the value of the attribute at the indicated location
     */
    public int getAttribute(int x, int y)
    {
        return _attributeData[y][x];
    }

// *******************************
// ***** METHODS - Rendering *****
// *******************************

    /**
     * renders the viewport portion of the playfield.
     *
     * @param g graphics context to render playfield into
     * @param comp component being rendered into
     * @param x x coordinate to render to
     * @param y y coordinate to renter to
     * @param viewport portion of the playfield to render
     */
    public synchronized void render(Graphics g, Component comp, int x, int y, Rectangle viewport)
    {
        int temp, offx, offy;
        int sskipx, sskipy;
        int stilex, stiley;
        int drawx, drawy;
        int eskipx, eskipy;

        temp = Math.abs(viewport.x % _tileWidth);
        if (viewport.x < 0)
        {
            temp = _tileWidth - temp;
            sskipx = Math.abs(viewport.x) / _tileWidth + 1;
            stilex = 0;
        }
        else
        {
            sskipx = 0;
            stilex = viewport.x / _tileWidth;
        }
        offx = x - temp;
        drawx = (viewport.width + temp + _tileWidth - 1) / _tileWidth;
        if ((stilex + drawx) >= _width)
        {
            eskipx = (stilex + drawx) - _width;
            drawx -= eskipx;
        }
        else
            eskipx = 0;

        temp = Math.abs(viewport.y % _tileHeight);
        if (viewport.y < 0)
        {
            temp = _tileHeight - temp;
            sskipy = Math.abs(viewport.y) / _tileHeight + 1;
            stiley = 0;
        }
        else
        {
            sskipy = 0;
            stiley = viewport.y / _tileHeight;
        }
        offy = y - temp;
        drawy = (viewport.height + temp + _tileHeight - 1) / _tileHeight;
        if ((stiley + drawy) >= _height)
        {
            eskipy = (stiley + drawy) - _height;
            drawy -= eskipy;
        }
        else
            eskipy = 0;
        
        int totalCols = sskipx + drawx + eskipx;
        int cntrCol, tempx = offx;
        int cntrRow, tempy = offy;
        for (cntrRow = 0; cntrRow < sskipy; ++cntrRow)
        {
            tempx = offx;
            for (cntrCol = 0; cntrCol < totalCols; ++cntrCol)
            {
                _tileset[0].render(g, comp, tempx, tempy);
                tempx += _tileWidth;
            }
            tempy += _tileHeight;
        }
        for (cntrRow = 0; cntrRow < drawy; ++cntrRow)
        {
            tempx = offx;
            for (cntrCol = 0; cntrCol < sskipx; ++cntrCol)
            {
                _tileset[0].render(g, comp, tempx, tempy);
                tempx += _tileWidth;
            }
            for (cntrCol = 0; cntrCol < drawx; ++cntrCol)
            {
try{
                _tileset[_tileData[stiley+cntrRow][stilex+cntrCol]].render(g, comp, tempx, tempy,_attributeData[stiley+cntrRow][stilex+cntrCol]);
                tempx += _tileWidth;
}catch(Exception e)
{
    System.out.println("Error:" + e.getMessage());
}
            }
            for (cntrCol = 0; cntrCol < eskipx; ++cntrCol)
            {
                _tileset[0].render(g, comp, tempx, tempy);
                tempx += _tileWidth;
            }
            tempy += _tileHeight;
        }
        for (cntrRow = 0; cntrRow < eskipy; ++cntrRow)
        {
            tempx = offx;
            for (cntrCol = 0; cntrCol < totalCols; ++cntrCol)
            {
                _tileset[0].render(g, comp, tempx, tempy);
                tempx += _tileWidth;
            }
            tempy += _tileHeight;
        }
    }

}
